/*
 * libnss_uto.c
 *
 * "Unknown To One", based on "All To One"
 *
 * /etc/libnss-uto will be searched for users, like usually /etc/passwd would be.
 * If the user is not found, the line with the username asterik "*" is used
 *
 * libnss_ato.c
 *
 * Nss module to map every requested user to a fixed one.
 * Ato stands for "All To One"
 *
 * Copyright (c) Pietro Donatini (pietro.donatini@unibo.it), 2007.
 *
 * this product may be distributed under the terms of
 * the GNU Lesser Public License.
 *
 * version 0.2
 *
 * CHANGELOG:
 * strip end of line in reading /etc/libnss-ato
 * suggested by Kyler Laird
 *
 * TODO:
 *
 * check bugs
 *
 */

#include <nss.h>
#include <pwd.h>
#include <shadow.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

/* #include <syslog.h> */

/* for security reasons */
#define MIN_UID_NUMBER   0
#define MIN_GID_NUMBER   0
#define CONF_FILE "/etc/libnss-uto.conf"

/*
 * The passwd structure in <pwd.h>
struct passwd
{
    char *pw_name;                Username.
    char *pw_passwd;              Password.
    __uid_t pw_uid;               User ID.
    __gid_t pw_gid;               Group ID.
    char *pw_gecos;               Real name.
    char *pw_dir;                 Home directory.
    char *pw_shell;               Shell program.
};

shadow passwd structure in <shadow.h>

struct spwd
{
    char *sp_namp;          Login name
    char *sp_pwdp;          Encrypted password
    long sp_lstchg;         Date of last change
    long sp_min;            Min #days between changes
    long sp_max;            Max #days between changes
    long sp_warn;           #days before pwd expires to warn user to change it
    long sp_inact;          #days after pwd expires until account is disabled
    long sp_expire;         #days since 1970-01-01 until account is disabled
    unsigned long sp_flag;  Reserved
};

*/

/*  What can be configured in /etc/libnss-uto.conf */
typedef struct uto_conf {
    char *pw_name;
    char *pw_gecos;
    char *pw_dir;
    char *pw_shell;
    __uid_t pw_uid;
    __gid_t pw_gid;
} uto_conf_t;

/* constructor */
uto_conf_t *
new_conf()
{
    uto_conf_t *conf;
    if ((conf = (uto_conf_t *) malloc(sizeof (uto_conf_t))) == NULL)
        return NULL;

    conf->pw_name  = NULL;
    conf->pw_gecos = NULL;
    conf->pw_dir   = NULL;
    conf->pw_shell = NULL;

    return conf;
}

/* destructor */
int
free_conf(uto_conf_t *conf)
{
    if ( conf->pw_name != NULL )
        free(conf->pw_name);
    if ( conf->pw_gecos != NULL )
        free(conf->pw_gecos);
    if ( conf->pw_dir != NULL )
        free(conf->pw_dir);
    if ( conf->pw_shell != NULL )
        free(conf->pw_shell);
    if ( conf != NULL )
        free(conf);
    return 0;
}

char *find_line(FILE *fd, const char *search_for, char *buff_delim, int column)
{
    char *buff_fgets = &buff_delim[1];
    buff_delim[0] = ':';

    while (fgets(buff_fgets, BUFSIZ, fd) != NULL) {
        if(buff_delim[strlen(buff_delim) - 1] == '\n')
            buff_delim[strlen(buff_delim) - 1] = '\0';
        if(buff_delim[strlen(buff_delim) - 1] == '\r')
            buff_delim[strlen(buff_delim) - 1] = '\0';
        strcat(buff_delim, ":");
        int idx = 0, cnt = 0;
        for (idx = 0; idx < strlen(buff_delim) ; ++idx) {
            if (buff_delim[idx] == ':') {
                if(cnt == column) {
                    if(strncmp(&buff_delim[idx], search_for, strlen(search_for)) == 0) {
                        return buff_fgets;
                    }
                }
                ++cnt;
            }
        }
    }
    return NULL;
}
/*
 * the configuration /etc/libnss-uto.conf has
 * the local user data as in /etc/passwd. For example:
 * dona:x:1001:1001:P D ,,,:/home/dona:/bin/bash
 * for unknown, or empty users the line with the username * is used:
 * *:x:1234:1234:Default User,,,:/tmp:/bin/bash
 *
 * this function can search for any column of that file, column counting starts at 0
 */

uto_conf_t *
read_conf(const char *key, int column)
{
    uto_conf_t *conf;
    FILE *fd;
    char buff[BUFSIZ + 2];        /* delimited buffer, e.g. there is a : at the start and the end */
    char *value, *b;
    char search_for[BUFSIZ];

    if (key == NULL || key[0] == '\0')
        return NULL;

    sprintf(search_for, ":%s:", key);

    if ((conf = new_conf()) == NULL)
        return NULL;

    if ((fd = fopen(CONF_FILE, "r")) == NULL ) {
        free_conf(conf);
        return NULL;
    }

    b = find_line(fd, search_for, buff, column);
    if ((b == NULL) && (column == 0)) {
        /* only re-search for the * username */
        rewind(fd);
        sprintf(search_for, ":*:");
        b = find_line(fd, search_for, buff, column);
    }

    fclose(fd);
    if (b == NULL) {
        free_conf(conf);
        return NULL;
    }

    /* we now have the correct line */

    /* pw_name */
    value = b;

    while (*b != ':' && *b != '\0')
        b++;

    if (*b != ':')
        goto format_error;

    *b = '\0';
    b++;

    conf->pw_name = strdup(value);

    /* NOT USED pw_passwd will be set equal to  x (we like shadows) */
    while (*b != ':' && *b != '\0')
        b++;

    if ( *b != ':' )
        goto format_error;

    b++;

    /* pw_uid */
    value = b;

    while (*b != ':' && *b != '\0')
        b++;

    if (*b != ':')
        goto format_error;

    b++;

    conf->pw_uid = atoi(value);

    if ( conf->pw_uid < MIN_UID_NUMBER )
                conf->pw_uid = MIN_UID_NUMBER;

    /* pw_gid */
    value = b;

    while (*b != ':' && *b != '\0')
        b++;

    if (*b != ':')
        goto format_error;

    b++;
    conf->pw_gid = atoi(value);
    if ( conf->pw_gid < MIN_GID_NUMBER )
        conf->pw_gid = MIN_GID_NUMBER;

    /* pw_gecos */
    value = b;

    while (*b != ':' && *b != '\0')
        b++;

    if (*b != ':')
        goto format_error;

    *b = '\0';
    b++;

    conf->pw_gecos = strdup (value);

    /* pw_dir */
    value = b;

    while (*b != ':' && *b != '\0')
        b++;

    if (*b != ':')
        goto format_error;

    *b = '\0';
    b++;

    conf->pw_dir = strdup (value);

    /* pw_shell takes the rest */
        /* Kyler Laird suggested to strip end line */
    value = b;
        /* lets get rid of the trailing : as well */
        while (*b != '\n' && *b != '\0' && *b != ':')
                b++;

        *b = '\0';

    conf->pw_shell = strdup(value);

    return conf;

    format_error:
        free (conf);
        return NULL;
}

/*
 * Allocate some space from the nss static buffer.  The buffer and buflen
 * are the pointers passed in by the C library to the _nss_ntdom_*
 * functions.
 *
 *  Taken fron glibc
 */

static char *
get_static(char **buffer, size_t *buflen, int len)
{
    char *result;

    /* Error check.  We return false if things aren't set up right, or
         * there isn't enough buffer space left. */

    if ((buffer == NULL) || (buflen == NULL) || (*buflen < len)) {
        return NULL;
    }

    /* Return an index into the static buffer */

    result = *buffer;
    *buffer += len;
    *buflen -= len;

    return result;
}

enum nss_status fill_passwd(uto_conf_t *conf,
                 const char *name,
                 struct passwd *p,
                 char *buffer,
                 size_t buflen,
                 int *errnop)
{
    if(name == NULL) {
        /* If out of memory */
        if ((p->pw_name = get_static(&buffer, &buflen, strlen(conf->pw_name) + 1)) == NULL) {
            return NSS_STATUS_TRYAGAIN;
        }
        strcpy(p->pw_name, conf->pw_name);
    } else {
        /* If out of memory */
        if ((p->pw_name = get_static(&buffer, &buflen, strlen(name) + 1)) == NULL) {
            return NSS_STATUS_TRYAGAIN;
        }

        /* pw_name stay as the name given */
        strcpy(p->pw_name, name);
    }

    if ((p->pw_passwd = get_static(&buffer, &buflen, strlen("x") + 1)) == NULL) {
                return NSS_STATUS_TRYAGAIN;
        }

    strcpy(p->pw_passwd, "x");

        p->pw_uid = conf->pw_uid; /* UID_NUMBER; */
        p->pw_gid = conf->pw_gid; /* GID_NUMBER; */

    if ((p->pw_gecos = get_static(&buffer, &buflen, strlen(conf->pw_gecos) + 1 )) == NULL) {
                return NSS_STATUS_TRYAGAIN;
        }

        strcpy(p->pw_gecos, conf->pw_gecos);

    if ((p->pw_dir = get_static(&buffer, &buflen, strlen(conf->pw_dir) + 1 )) == NULL) {
                return NSS_STATUS_TRYAGAIN;
        }

        strcpy(p->pw_dir, conf->pw_dir);

    if ((p->pw_shell = get_static(&buffer, &buflen, strlen(conf->pw_shell) + 1 )) == NULL) {
                return NSS_STATUS_TRYAGAIN;
        }

        strcpy(p->pw_shell, conf->pw_shell);

        free_conf(conf);

    return NSS_STATUS_SUCCESS;
}

enum nss_status
_nss_uto_getpwnam_r( const char *name,
             struct passwd *p,
                 char *buffer,
                 size_t buflen,
                 int *errnop)
{
    uto_conf_t *conf;
    if ((conf = read_conf(name, 0)) == NULL) {
        return NSS_STATUS_NOTFOUND;
    }

    return fill_passwd(conf, name, p, buffer, buflen, errnop);
}

enum nss_status
_nss_uto_getspnam_r( const char *name,
                     struct spwd *s,
                     char *buffer,
                     size_t buflen,
                     int *errnop)
{
        /* If out of memory */
        if ((s->sp_namp = get_static(&buffer, &buflen, strlen(name) + 1)) == NULL) {
                return NSS_STATUS_TRYAGAIN;
        }

        strcpy(s->sp_namp, name);

        if ((s->sp_pwdp = get_static(&buffer, &buflen, strlen("!") + 1)) == NULL) {
                return NSS_STATUS_TRYAGAIN;
        }

        strcpy(s->sp_pwdp, "!");

        s->sp_lstchg = 13571;
        s->sp_min    = 0;
        s->sp_max    = 99999;
        s->sp_warn   = 7;

        return NSS_STATUS_SUCCESS;
}

enum nss_status
_nss_uto_getpwuid_r( const __uid_t uid,
             struct passwd *p,
                 char *buffer,
                 size_t buflen,
                 int *errnop)
{
    uto_conf_t *conf;
    char c_uid[8];
    sprintf(c_uid, "%u", uid);
    if ((conf = read_conf(c_uid, 2)) == NULL) {
        return NSS_STATUS_NOTFOUND;
    }

    if(conf->pw_name[0] == '\0')
        return fill_passwd(conf, "default", p, buffer, buflen, errnop);
    else
        return fill_passwd(conf, NULL, p, buffer, buflen, errnop);
}
